#!/usr/bin/env perl

# This script automates the process of generating a set of evolved
# sequences using specific evolution parameters for seq-gen and
# sgEvolver, aligning those sequences, and scoring the alignments


use strict;
use POSIX;
require simujobparams;

if( @ARGV != 1 ){
	die "Usage: simujobrun.pl <mauve|mavid|mlagan|none>";
}

# Set the location of the evolution and scoring tools here:
# this is the directory where seq-gen, sgEvolver, and scoreAlignment reside
my $tools_dir="/home/koadman/bin/";
my $mauve_dir="/u/d/a/darling/mauve/";
my $mavid_dir="/u/d/a/darling/mavid/";
my $lagan_dir="/u/d/a/darling/lagan121/";
my $xmfa_alignment = "";

# an array of files that need to be deleted when cleaning up
my @delete_files = ();

# ROADMAP:
# 1) simulate evolution according to parameters
# 2) align genomes using the selected aligner
# 3) score alignments

#
# read in the job specification
#

my $job_line;

		# write command lines executed to a file
open( COMLINES, ">command_lines.txt"  );


# generate an ancestral base
open( ANCESTRAL, ">$simujobparams::ancestral_seq_name" );
print ANCESTRAL " 1 $simujobparams::seq_length\n";
print ANCESTRAL "Ancestral\n";
close ANCESTRAL;
	
my $rval;

		# extract the specified ancestral sequence
my $extract_cl = "dd if=$simujobparams::ancestral_donor bs=1 skip=$simujobparams::ancestral_start count=$simujobparams::seq_length";
executeCommand( $extract_cl, ">$simujobparams::ancestral_seq_name", "ancestral_dd.err" );

#my $simujobparams::sgEvolver_random = int(rand(0x7FFFFFFF));

		# generate a data set
my $seqgen_cl = $tools_dir."seq-gen -mHKY -a $simujobparams::gamma_shape -i 0 -z $simujobparams::seqgen_random -f ";
$seqgen_cl .= "$simujobparams::nt_a_freq $simujobparams::nt_c_freq $simujobparams::nt_g_freq $simujobparams::nt_t_freq ";
$seqgen_cl .= "-t .5 -k 1 -wa -on -s $simujobparams::nt_sub_scale $simujobparams::tree_filename ";
$seqgen_cl .= "< $simujobparams::ancestral_seq_name ";
executeCommand( $seqgen_cl, ">$simujobparams::seqgen_out_name", "seqgen_ancestral.err" );

		# generate a donor base
open( DONOR, ">$simujobparams::donor_seq_name" );
print DONOR " 1 $simujobparams::seq_length\n";
print DONOR "Donor\n";
close DONOR;


		# extract the specified donor sequence
$extract_cl = "dd if=$simujobparams::ancestral_donor bs=1 skip=$simujobparams::donor_start count=$simujobparams::seq_length";
executeCommand( $extract_cl, ">$simujobparams::donor_seq_name", "donor_dd.err" );

		# generate a donor data set
$seqgen_cl = $tools_dir."seq-gen -mHKY -a $simujobparams::gamma_shape -i 0 -z $simujobparams::seqgen_donor_random  -f ";
$seqgen_cl .= "$simujobparams::nt_a_freq $simujobparams::nt_c_freq $simujobparams::nt_g_freq $simujobparams::nt_t_freq ";
$seqgen_cl .= "-t .5 -k 1 -wa -on -s $simujobparams::nt_sub_scale $simujobparams::tree_filename ";
$seqgen_cl .= "< $simujobparams::donor_seq_name ";
executeCommand( $seqgen_cl, ">$simujobparams::seqgen_out_name", "seqgen_donor.err" );


		# continue evolution with sgEvolver
my $sgevolver_cl = "time ".$tools_dir."sgEvolver --indel-freq=".$simujobparams::indel_rate." --small-ht-freq=$simujobparams::small_ht_rate --small-ht-size=$simujobparams::small_ht_size".
"--large-ht-freq=$simujobparams::large_ht_rate --inversion-freq=$simujobparams::inv_rate --large-ht-min=$simujobparams::large_ht_min --large-ht-max=$simujobparams::large_ht_max".
"--random-seed=$simujobparams::sgEvolver_random".
" --inversion-size=$simujobparams::inv_size $simujobparams::tree_filename $simujobparams::seqgen_out_name $simujobparams::evolved_seqs_name $simujobparams::evolved_seqs_fname";
$rval = executeCommand( $sgevolver_cl, "sgEvolver.out", "sgEvolver.err" );

# die if sgEvolver failed
die "Failure in sgEvolver" if( $rval != 0 );

die "Failure in sgEvolver" unless -e "$simujobparams::evolved_seqs_name";
die "Failure in sgEvolver" unless -e "$simujobparams::evolved_seqs_fname";

# delete seq-gen related files
push( @delete_files, $simujobparams::seqgen_out_name );
push( @delete_files, $simujobparams::ancestral_seq_name );
push( @delete_files, $simujobparams::donor_seq_name );

deleteFiles( @delete_files );

#
# STEP 2:  align genomes using the selected aligner
#

alignMauve() if( $ARGV[0] eq "mauve" );
alignMavid() if( $ARGV[0] eq "mavid" );
alignMlagan() if( $ARGV[0] eq "mlagan" );
alignSlagan() if( $ARGV[0] eq "slagan" );
exit(0) if( $ARGV[0] eq "none" );

#
# STEP 3: score the alignment
#

my $scoreAlignment_path = $tools_dir."scoreAlignment";

my $score_cl = "$scoreAlignment_path $simujobparams::evolved_seqs_name $xmfa_alignment $simujobparams::evolved_seqs_fname";
executeCommand( $score_cl, "scores.txt", "scoring.err" );

# do a backbone scoring
my $bb_file = "backbone.dat";
push( @delete_files, $bb_file );
my $extractbb_cl = $tools_dir."extractBackbone $simujobparams::evolved_seqs_fname $simujobparams::evolved_seqs_name 50 50 $bb_file";
executeCommand( $extractbb_cl, "extract_bb.out", "extract_bb.err" );

$score_cl = "$scoreAlignment_path $bb_file $xmfa_alignment $simujobparams::evolved_seqs_fname";
executeCommand( $score_cl, "bb_scores.txt", "bb_scoring.err" );

#
# delete alignment files
#
#system( "rm $simujobparams::evolved_seqs_name" );
#system( "rm $simujobparams::evolved_seqs_fname" );

push( @delete_files, $simujobparams::evolved_seqs_name );
push( @delete_files, $simujobparams::evolved_seqs_fname );

deleteFiles( @delete_files );
exit(0);

sub executeCommand {
  my $command = shift;
  my $stdout_file = shift;
  my $stderr_file = shift;
  $command .= " >$stdout_file 2>$stderr_file";
  print "Executing $command\n";
  print COMLINES "$command\n";
  my $rval = system($command);
  `echo "Exited with code $rval" >> $stderr_file` if $rval != 0;
  return $rval;
}

sub deleteFiles {
  foreach(@_){
    `rm -f $_`;
    print COMLINES "rm -f $_\n";
  }
}

sub alignMauve {
  $xmfa_alignment = "aligned.dat";
  push( @delete_files, $xmfa_alignment );
  push( @delete_files, "aligned.mums" );
  push( @delete_files, "aligned.mauve" );
  my $mauve_cl = $mauve_dir."mauveAligner --mums -o aligned.mums $simujobparams::evolved_seqs_fname";
  executeCommand( $mauve_cl, "mums.out", "mums.err" );

  $mauve_cl = $mauve_dir."mauveAligner -i aligned.mums -o aligned.mauve ".
  "--output-alignment=$xmfa_alignment $simujobparams::evolved_seqs_fname";
  $rval = executeCommand( $mauve_cl, "aligner.out", "aligner.err" );

  # die if mauve failed
  die "Failure in aligner" if( $rval != 0 );
}

sub alignMavid {
  $xmfa_alignment = "mavid.dat";

  # dust the sequence for mavid
  my $dust_cl = "$mavid_dir/mdust $simujobparams::evolved_seqs_fname";
  executeCommand( $dust_cl, "$simujobparams::evolved_seqs_fname.masked", "mdust.err" );

  my $mavid_cl = $mavid_dir."/mavid $simujobparams::tree_filename $simujobparams::evolved_seqs_fname";
  executeCommand( $mavid_cl, "mavid.out", "mavid.err" );


  # step 2.1 convert mavid.mfa to xmfa format
  my $xmfa_cl = "$tools_dir/mfa2xmfa mavid.mfa $xmfa_alignment";
  executeCommand( $xmfa_cl, "mfa2xmfa.out", "mfa2xmfa.err" );

  push( @delete_files, "$simujobparams::evolved_seqs_fname.masked" );
  push( @delete_files, "mavid.mfa" );
  push( @delete_files, "mavid.phy" );
  push( @delete_files, "$xmfa_alignment" );

}


sub alignMlagan {

}

sub alignSlagan {

}

